package affuse

import (
	"context"
	"io"
	"os"
	"sync"

	"bazil.org/fuse"
	"bazil.org/fuse/fs"
	"github.com/sirupsen/logrus"
	"github.com/spf13/afero"
)

type File interface {
	fs.Node
	fs.Handle
	fs.NodeOpener
	fs.HandleReleaser
	fs.HandleReader
	fs.HandleWriter
	fs.HandleFlusher
	fs.NodeSetattrer
	fs.NodeFsyncer
}

var _ File = &file{}

type file struct {
	fs afero.Fs
	file afero.File
	mu sync.RWMutex
	count uint64
}

func (f *file) Attr(ctx context.Context, attr *fuse.Attr) error {
	logrus.Debugf("Attr %s", f.file.Name())
	s, err := f.file.Stat()
	if err != nil {
		return err
	}
	attrs(attr, s)
	return nil
}

func (f *file) Open(ctx context.Context, req *fuse.OpenRequest, resp *fuse.OpenResponse) (fs.Handle, error) {
	logrus.Debugf("Open %s (%s)", f.file.Name(),req.Flags)
	if req.Flags.IsReadOnly() {
		// we don't need to track read-only handles
		return f, nil
	}
	f.mu.Lock()
	defer f.mu.Unlock()
	if f.count != 0 {
		f.count ++
		logrus.Debugf("Open %s count: %d", f.file.Name(), f.count)
		return f, nil
	}
	var err error
	f.file, err = f.fs.OpenFile(f.file.Name(), int(req.Flags), os.ModePerm)
	if err != nil {
		logrus.Errorf("Open %s: %v", f.file.Name(), err)
		return nil, err
	}
	f.count++
	logrus.Debugf("Open %s count: %d", f.file.Name(), f.count)
	return f, nil
}

func (f *file) Release(ctx context.Context, req *fuse.ReleaseRequest) error {
	logrus.Debugf("Release %s", f.file.Name())
	if req.Flags.IsReadOnly() {
		// we don't need to track read-only handles
		return nil
	}

	f.mu.Lock()
	defer f.mu.Unlock()
	f.count--
	logrus.Debugf("Release %s count: %d", f.file.Name(), f.count)
	if f.count != 0 {
		return nil
	}
	if err := f.file.Close(); err != nil {
		return err
	}
	return nil
}

func (f *file) Read(ctx context.Context, req *fuse.ReadRequest, resp *fuse.ReadResponse) error {
	logrus.Debugf("Read %s", f.file.Name())
	f.mu.RLock()
	defer f.mu.RUnlock()
	buf := make([]byte, req.Size)
	n, err := f.file.ReadAt(buf, req.Offset)
	if err == io.EOF {
		return nil
	}
	if err != nil {
		logrus.Errorf("Read %s: %v", f.file.Name(), err)
		return err
	}
	resp.Data = buf[:n]
	return nil
}

func (f *file) Write(ctx context.Context, req *fuse.WriteRequest, resp *fuse.WriteResponse) error {
	s, _ := f.file.Stat()
	logrus.Debugf("Write %s (mod time: %v)", f.file.Name(), s.ModTime())
	f.mu.Lock()
	defer f.mu.Unlock()
	n, err := f.file.WriteAt(req.Data, req.Offset)
	if err != nil {
		logrus.Errorf("Write %s: %v", f.file.Name(), err)
		return err
	}
	resp.Size = n
	s, _ = f.file.Stat()
	logrus.Debugf("End Write %s (mod time: %v)", f.file.Name(), s.ModTime())
	return nil
}

func (f *file) Flush(ctx context.Context, req *fuse.FlushRequest) error {
	logrus.Debugf("Flush %s", f.file.Name())
	return nil
}

func (f *file) Setattr(ctx context.Context, req *fuse.SetattrRequest, resp *fuse.SetattrResponse) error {
	logrus.Debugf("Setattr %s", f.file.Name())
	s, err := f.file.Stat()
	if err != nil {
		return err
	}
	if req.Valid.Size() {
		logrus.Debugf("Setattr %s size: %d", f.file.Name(), req.Size)
		newLen := int64(req.Size)
		switch {
		case newLen > s.Size():
			if _, err := f.file.WriteAt(make([]byte, newLen-s.Size()), s.Size()); err != nil {
				return err
			}
		case newLen < s.Size():
			if err := f.file.Truncate(s.Size() - newLen); err != nil {
				return err
			}
		}
	}
	if req.Valid.Mode() {
		logrus.Debugf("Setattr %s mod: %d", f.file.Name(), req.Mode)
		if err := f.fs.Chmod(f.file.Name(), req.Mode); err != nil {
			return err
		}
	}
	s, err = f.file.Stat()
	if err != nil {
		return err
	}
	attrs(&resp.Attr, s)
	return nil
}

func (f *file) Fsync(ctx context.Context, req *fuse.FsyncRequest) error {
	return nil
}

