package affuse

import (
	"context"
	"errors"
	"fmt"
	"sync"

	"bazil.org/fuse"
	"bazil.org/fuse/fs"
	"github.com/sirupsen/logrus"
	"github.com/spf13/afero"
)

type FS interface {
	fs.FS
	Mount(ctx context.Context, mountPoint string) error
	Unmount() error
}

func New(afs afero.Fs) (FS, error) {
	if afs == nil {
		return nil, errors.New("afs cannot be nil")
	}
	root, err := afs.Open("/")
	if err != nil {
		return nil, fmt.Errorf("failed to open file system root: %v", err)
	}
	s, err := root.Stat()
	if err != nil {
		return nil, fmt.Errorf("failed get file system root stats: %v", err)
	}
	if !s.IsDir() {
		return nil, errors.New("file system root is not a directory")
	}
	fuse.MaxReadahead(128*1024)
	return &aferoFs{fs: afs, root: root}, nil
}

type aferoFs struct {
	fs   afero.Fs
	root afero.File
	mu sync.Mutex
	mount string
}

func (a *aferoFs) Root() (fs.Node, error) {
	logrus.Debugf("Root: %s", a.root.Name())
	return &dir{
		fs:   a.fs,
		file: a.root,
	}, nil
}

func (a *aferoFs) Mount(ctx context.Context, p string) error {
	a.mu.Lock()
	c, err := fuse.Mount(p)
	if err != nil {
		a.mu.Unlock()
		return err
	}
	defer c.Close()
	a.mount = p
	a.mu.Unlock()
	go func() {
		for {
			select {
			case <- ctx.Done():
				a.Unmount()
			}
		}
	}()
	if err := fs.Serve(c, a); err != nil {
		return err
	}
	// check if the mount process has an error to report
	<-c.Ready
	if err := c.MountError; err != nil {
		return err
	}
	return nil
}

func (a *aferoFs) Unmount() error {
	a.mu.Lock()
	defer a.mu.Unlock()
	return fuse.Unmount(a.mount)
}
